"""Route models for MetroBus API."""
from __future__ import annotations

from dataclasses import dataclass, field
from typing import TypedDict


class RouteData(TypedDict):
    """Route data for MetroBus API."""

    RouteID: str
    Name: str
    LineDescription: list[str]


@dataclass
class Route:
    """A MetroBus Route."""

    data: RouteData
    route_id: str = field(init=False)
    name: str = field(init=False)
    line_description: list[str] = field(init=False)

    def __post_init__(self) -> None:
        """Post init."""
        self.route_id = self.data["RouteID"]
        self.name = self.data["Name"]
        self.line_description = self.data["LineDescription"]



"""NextTrain models for MetroRail WMATA API."""
from __future__ import annotations

from contextlib import suppress
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, TypedDict

if TYPE_CHECKING:
    from .. import MetroRail
    from .line import Line
    from .station import Station


class NextTrainData(TypedDict):
    """NextTrain data for MetroRail WMATA API."""

    Car: int | str | None
    Destination: str
    DestinationCode: str | None
    DestinationName: str | None
    Group: str
    Line: str
    LocationCode: str
    LocationName: str
    Min: int | str | None


@dataclass
class NextTrain:
    """NextTrain."""

    bus: "MetroRail"
    data: NextTrainData
    num_cars: int | None = field(init=False)
    destination: str = field(init=False)
    destination_station_code: str | None = field(init=False)
    destination_station_name: str | None = field(init=False)
    group: int = field(init=False)
    line_code: str | None = field(init=False)
    location_code: str = field(init=False)
    location_name: str = field(init=False)
    minutes: int | str | None = field(init=False)

    def __post_init__(self) -> None:
        """Post init."""
        if isinstance((car := self.data["Car"]), int) and car != 0:
            self.num_cars = int(car)
        else:
            self.num_cars = None
        self.destination = self.data["Destination"]
        self.destination_station_code = self.data["DestinationCode"]
        self.destination_station_name = self.data["DestinationName"]
        self.group = int(self.data["Group"])
        if (line_code := self.data["Line"]) in (None, "No"):
            self.line_code = None
        else:
            self.line_code = line_code
        self.location_code = self.data["LocationCode"]
        self.location_name = self.data["LocationName"]
        if (minutes := self.data["Min"]) in ("-", "---") or minutes is None:
            self.minutes = None
        else:
            self.minutes = minutes
            with suppress(ValueError):
                self.minutes = int(minutes)

    @property
    def destination_station(self) -> "Station" | None:
        """Return the destination Station."""
        if not self.destination_station_code:
            return None
        return self.bus.stations[self.destination_station_code]

    @property
    def line(self) -> "Line" | None:
        """Return the Line."""
        if not self.line_code:
            return None
        return self.bus.lines[self.line_code]

    @property
    def location(self) -> "Station":
        """Return the location Station."""
        return self.bus.stations[self.location_code]



"""Station timings models for MetroRail WMATA API."""
from __future__ import annotations

from dataclasses import dataclass, field
from datetime import date, datetime, time
from typing import TYPE_CHECKING, TypedDict

from ...const import TZ

if TYPE_CHECKING:
    from .. import MetroRail
    from .station import Station


class TrainTimingData(TypedDict):
    """Train timing data for MetroRail WMATA API."""

    Time: str
    DestinationStation: str


@dataclass
class TrainTiming:
    """Train timing for MetroRail WMATA API."""

    bus: "MetroRail"
    data: TrainTimingData
    time: time = field(init=False)
    destination_station_code: str = field(init=False)

    def __post_init__(self) -> None:
        """Post init."""
        self.time = time.fromisoformat(self.data["Time"]).replace(tzinfo=TZ)
        self.destination_station_code = self.data["DestinationStation"]

    @property
    def destination_station(self) -> "Station":
        """Return the destination station."""
        return self.bus.stations[self.destination_station_code]


class DayTimeData(TypedDict):
    """Day time data for MetroRail WMATA API."""

    OpeningTime: str
    FirstTrains: list[TrainTimingData]
    LastTrains: list[TrainTimingData]


@dataclass
class DayTime:
    """Day time for MetroRail WMATA API."""

    bus: "MetroRail"
    data: DayTimeData
    day: str
    day_of_week: int
    opening_time: time = field(init=False)
    first_trains: list[TrainTiming] = field(init=False)
    last_trains: list[TrainTiming] = field(init=False)

    def __post_init__(self) -> None:
        """Post init."""
        self.opening_time = time.fromisoformat(self.data["OpeningTime"]).replace(
            tzinfo=TZ
        )
        self.first_trains = [
            TrainTiming(self.bus, train) for train in self.data["FirstTrains"]
        ]
        self.last_trains = [
            TrainTiming(self.bus, train) for train in self.data["LastTrains"]
        ]


class StationTimeData(TypedDict):
    """Station time data for MetroRail WMATA API."""

    Code: str
    StationName: str
    Monday: DayTimeData
    Tuesday: DayTimeData
    Wednesday: DayTimeData
    Thursday: DayTimeData
    Friday: DayTimeData
    Saturday: DayTimeData
    Sunday: DayTimeData


@dataclass
class StationTime:
    """Station time for MetroRail WMATA API."""

    bus: "MetroRail"
    data: StationTimeData
    station_code: str = field(init=False)
    name: str = field(init=False)
    days_of_week: dict[int, DayTime] = field(init=False)

    def __post_init__(self) -> None:
        """Post init."""
        self.station_code = self.data["Code"]
        self.name = self.data["StationName"]
        self.days_of_week = {
            1: DayTime(self.bus, self.data["Monday"], "Monday", 1),
            2: DayTime(self.bus, self.data["Tuesday"], "Tuesday", 2),
            3: DayTime(self.bus, self.data["Wednesday"], "Wednesday", 3),
            4: DayTime(self.bus, self.data["Thursday"], "Thursday", 4),
            5: DayTime(self.bus, self.data["Friday"], "Friday", 5),
            6: DayTime(self.bus, self.data["Saturday"], "Saturday", 6),
            7: DayTime(self.bus, self.data["Sunday"], "Sunday", 7),
        }

    @property
    def station(self) -> "Station":
        """Return the station."""
        return self.bus.stations[self.station_code]

    def get_day_time(self, date_: datetime | date) -> DayTime:
        """Return the day time for the given date."""
        return self.days_of_week[date_.isoweekday()]



"""Module for interacting with MetroBus API."""
from __future__ import annotations

from datetime import date
from typing import TYPE_CHECKING, Any, cast

from ..models.area import Area
from .const import BusEndpoint
from .models.bus_incident import BusIncident
from .models.live_position import LiveBusPosition
from .models.next_bus import NextBus
from .models.route import Route
from .models.route_path import RoutePath, RoutePathData
from .models.route_schedule import RouteSchedule, RouteScheduleData
from .models.stop import Stop, StopData
from .models.stop_schedule import StopArrival

if TYPE_CHECKING:
    from ..client import Client


class MetroBus:
    """Class to represent client for MetroBus APIs."""

    client: "Client"
    routes: dict[str, Route]
    stops: dict[str, Stop]

    def __init__(self, client: "Client") -> None:
        """Initialize."""
        self.client = client
        self.routes = {}
        self.stops = {}

    async def load_data(self) -> None:
        """Load the base data."""
        self.routes = await self.get_all_routes()
        self.stops = await self.get_stops()

    def get_stop_from_stop_data(
        self, stop_data: StopData, use_internal_data: bool = True
    ) -> Stop:
        """Get stop from stop ID."""
        stop_id = stop_data["StopID"]
        if not isinstance(stop_id, str):
            stop_id = stop_data["Name"]
        if not use_internal_data:
            return Stop(self, stop_data)
        if stop_id not in self.stops:
            raise ValueError(f"Stop `{stop_id}` not found.")
        return self.stops[stop_id]

    def get_all_stops_from_stop_data(
        self, stops_data: list[StopData], use_internal_data: bool = True
    ) -> dict[str, Stop]:
        """Get all stops from list of StopData."""
        data: dict[str, Stop] = {}
        for stop_data in stops_data:
            stop = self.get_stop_from_stop_data(stop_data, use_internal_data)
            data[stop.stop_id] = stop

        return data

    async def get_all_routes(self) -> dict[str, Route]:
        """Get all routes."""
        data = await self.client.fetch(BusEndpoint.ROUTES)
        return {
            route_data["RouteID"]: Route(route_data) for route_data in data["Routes"]
        }

    async def get_stops(self, area: Area | None = None) -> dict[str, Stop]:
        """Get stops."""
        params: dict | None = None
        if area:
            params = dict(area.to_dict())
        data = await self.client.fetch(BusEndpoint.STOPS, params=params)
        return self.get_all_stops_from_stop_data(data["Stops"], use_internal_data=False)

    async def get_live_positions(
        self,
        route: Route | None = None,
        area: Area | None = None,
    ) -> list[LiveBusPosition]:
        """
        Get live bus position(s).

        If no route or area is provided, all bus positions will be
        returned.
        """
        params: dict[str, Any] = {}
        if route:
            params["RouteID"] = route.route_id
        if area:
            params.update(area.to_dict())

        data = await self.client.fetch(BusEndpoint.POSITIONS, params=params)
        return sorted(
            [LiveBusPosition(self, position) for position in data["BusPositions"]],
            key=lambda position: position.trip_start_time,
        )

    async def get_bus_incidents(self, route: Route | None = None) -> list[BusIncident]:
        """
        Get bus incidents.

        If no route is provided, all bus incidents for all routes will be returned.
        """
        params = None
        if route:
            params = {"Route": route.route_id}

        data = await self.client.fetch(BusEndpoint.BUS_INCIDENTS, params=params)

        return sorted(
            [BusIncident(self, incident) for incident in data["BusIncidents"]],
            reverse=True,
            key=lambda incident: incident.date_updated,
        )

    async def get_route_path(
        self, route: Route, date_: date | None = None
    ) -> RoutePath:
        """
        Return path for a given route.

        If no date_ is provided, the current date will be used.
        """
        params = {"RouteID": route.route_id}
        if date_:
            params["Date"] = date_.strftime("%Y-%m-%d")

        data = cast(
            RoutePathData,
            await self.client.fetch(BusEndpoint.ROUTE_PATH, params=params),
        )
        return RoutePath(self, data)

    async def get_route_schedule(
        self,
        route: Route,
        date_: date | None = None,
        include_variations: bool = False,
    ) -> RouteSchedule:
        """
        Return schedule for a given route.

        - If no date_ is provided, the current date will be used.
        - include_variations specifies whether variations for the input route should be
        included.
        """
        params = {
            "RouteID": route.route_id,
            "IncludingVariations": str(include_variations).lower(),
        }
        if date_:
            params["Date"] = date_.strftime("%Y-%m-%d")

        data = cast(
            RouteScheduleData,
            await self.client.fetch(BusEndpoint.ROUTE_SCHEDULE, params=params),
        )
        return RouteSchedule(self, route, data)

    async def get_next_buses_at_stop(self, stop: Stop) -> list[NextBus]:
        """Return next buses for a given stop."""
        data = await self.client.fetch(
            BusEndpoint.NEXT_BUSES, params={"StopID": stop.stop_id}
        )
        return [NextBus(self, next_bus_data) for next_bus_data in data["Predictions"]]

    async def get_stop_schedule(
        self, stop: Stop, date_: date | None = None
    ) -> list[StopArrival]:
        """
        Return schedule for a given stop.

        If no date_ is provided, the current date will be used.
        """
        params = {"StopID": stop.stop_id}
        if date_:
            params["Date"] = date_.strftime("%Y-%m-%d")

        data = await self.client.fetch(BusEndpoint.STOP_SCHEDULE, params=params)
        return sorted(
            [
                StopArrival(self, stop, stop_arrival_schedule)
                for stop_arrival_schedule in data["ScheduleArrivals"]
            ],
            key=lambda stop_arrival_schedule: stop_arrival_schedule.schedule_time,
        )



"""Constants for WMATA API."""
from zoneinfo import ZoneInfo

BASE_URL = "https://api.wmata.com"
TZ = ZoneInfo("US/Eastern")
HEADER_PREFIX = "wmataio"
CLASS_HEADER = f"{HEADER_PREFIX}-class"
ENUM_HEADER = f"{HEADER_PREFIX}-enum"
ADDITIONAL_PATH_HEADER = f"{HEADER_PREFIX}-additional-path"



"""Train position models for MetroRail API."""
from __future__ import annotations

from dataclasses import dataclass, field
from enum import IntEnum
from typing import TYPE_CHECKING, Literal, TypedDict

if TYPE_CHECKING:
    from .. import MetroRail
    from .line import Line
    from .station import Station


class TrainDirection(IntEnum):
    """Enum for all known train directions."""

    NORTHBOUND_OR_EASTBOUND = 1
    SOUTHBOUND_OR_WESTBOUND = 2


class LiveTrainPositionData(TypedDict):
    """Train position data for MetroRail API."""

    CarCount: int
    CircuitId: int
    DestinationStationCode: str | None
    DirectionNum: int
    LineCode: str
    SecondsAtLocation: int
    ServiceType: Literal["NoPassengers", "Normal", "Special", "Unknown"]
    TrainId: str
    TrainNumber: str


@dataclass
class LiveTrainPosition:
    """A MetroRail Bus LiveTrainPosition."""

    bus: "MetroRail"
    data: LiveTrainPositionData
    car_count: int = field(init=False)
    circuit_id: int = field(init=False)
    destination_station_code: str | None = field(init=False)
    direction: TrainDirection = field(init=False)
    line_code: str = field(init=False)
    seconds_at_location: int = field(init=False)
    service_type: Literal["NoPassengers", "Normal", "Special", "Unknown"] = field(
        init=False
    )
    train_id: str = field(init=False)
    train_number: str = field(init=False)

    def __post_init__(self) -> None:
        """Post init."""
        self.car_count = self.data["CarCount"]
        self.circuit_id = self.data["CircuitId"]
        self.destination_station_code = self.data["DestinationStationCode"]
        self.direction = TrainDirection(self.data["DirectionNum"])
        self.line_code = self.data["LineCode"]
        self.seconds_at_location = self.data["SecondsAtLocation"]
        self.service_type = self.data["ServiceType"]
        self.train_id = self.data["TrainId"]
        self.train_number = self.data["TrainNumber"]

    @property
    def destination_station(self) -> "Station" | None:
        """Return the destination station."""
        if not self.destination_station_code:
            return None
        return self.bus.stations[self.destination_station_code]

    @property
    def line(self) -> "Line" | None:
        """Return the line."""
        if not self.line_code:
            return None
        return self.bus.lines[self.line_code]



"""Constants for MetroRail API."""
from __future__ import annotations

from enum import Enum

from ..const import BASE_URL


class RailEndpoint(Enum):
    """Endpoints for all MetroRail endpoints."""

    ELEVATOR_ESCALATOR_INCIDENTS = f"{BASE_URL}/Incidents.svc/json/ElevatorIncidents"
    RAIL_INCIDENTS = f"{BASE_URL}/Incidents.svc/json/Incidents"
    LINES = f"{BASE_URL}/Rail.svc/json/jLines"
    NEXT_TRAINS = f"{BASE_URL}/StationNextTrain.svc/json/GetNextTrain"
    STANDARD_ROUTES = f"{BASE_URL}/TrainPositions/StandardRoutes"
    STATION_ENTRANCES = f"{BASE_URL}/Rail.svc/json/jStationEntrances"
    STATION_INFORMATION = f"{BASE_URL}/Rail.svc/json/jStationInfo"
    STATION_PARKING_INFORMATION = f"{BASE_URL}/Rail.svc/json/jStationParking"
    STATION_TIMINGS = f"{BASE_URL}/Rail.svc/json/jStationTimes"
    STATION_TO_STATION_INFO = f"{BASE_URL}/Rail.svc/json/jSrcStationToDstStationInfo"
    STATION_TO_STATION_PATH = f"{BASE_URL}/Rail.svc/json/jPath"
    STATIONS = f"{BASE_URL}/Rail.svc/json/jStations"
    TRACK_CIRCUITS = f"{BASE_URL}/TrainPositions/TrackCircuits"
    TRAIN_POSITIONS = f"{BASE_URL}/TrainPositions/TrainPositions"



"""Line models for MetroRail WMATA API."""
from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING, TypedDict

if TYPE_CHECKING:
    from .. import MetroRail
    from .standard_route import StandardRoute
    from .station import Station


class LineData(TypedDict):
    """Line data for MetroRail WMATA API."""

    DisplayName: str
    StartStationCode: str
    EndStationCode: str
    InternalDestination1: str
    InternalDestination2: str
    LineCode: str


@dataclass
class Line:
    """MetroRail Line."""

    bus: "MetroRail"
    data: LineData
    standard_routes: list["StandardRoute"]
    display_name: str = field(init=False)
    start_station_code: str = field(init=False)
    end_station_code: str = field(init=False)
    internal_destination_code_1: str | None = field(init=False, default=None)
    internal_destination_code_2: str | None = field(init=False, default=None)
    line_code: str = field(init=False)

    def __post_init__(self) -> None:
        """Post init."""
        self.line_code = self.data["LineCode"]
        self.display_name = self.data["DisplayName"]
        self.start_station_code = self.data["StartStationCode"]
        self.end_station_code = self.data["EndStationCode"]
        if internal_destination_1 := self.data["InternalDestination1"]:
            self.internal_destination_code_1 = internal_destination_1
        if internal_destination_2 := self.data["InternalDestination2"]:
            self.internal_destination_code_2 = internal_destination_2

    @property
    def start_station(self) -> "Station":
        """Return the start station."""
        return self.bus.stations[self.start_station_code]

    @property
    def end_station(self) -> "Station":
        """Return the end station."""
        return self.bus.stations[self.end_station_code]

    @property
    def internal_destination_1(self) -> "Station" | None:
        """Return the internal destination 1."""
        if not self.internal_destination_code_1:
            return None
        return self.bus.stations[self.internal_destination_code_1]

    @property
    def internal_destination_2(self) -> "Station" | None:
        """Return the internal destination 2."""
        if not self.internal_destination_code_2:
            return None
        return self.bus.stations[self.internal_destination_code_2]



"""Stop models for MetroBus API."""
from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING, TypedDict

from ...models.coordinates import Coordinates

if TYPE_CHECKING:
    from .. import MetroBus
    from .route import Route


class StopData(TypedDict, total=False):
    """Stop data for MetroBus API."""

    StopID: str | int | None
    Lat: float
    Lon: float
    Name: str
    Routes: list[str]


@dataclass
class Stop:
    """A MetroBus Stop."""

    bus: "MetroBus"
    data: StopData
    stop_id: str = field(init=False)
    coordinates: Coordinates = field(init=False)
    name: str = field(init=False)
    route_ids: list[str] = field(init=False)

    def __post_init__(self) -> None:
        """Post init."""
        if isinstance(self.data["StopID"], str):
            self.stop_id = self.data["StopID"]
        else:
            self.stop_id = self.data["Name"]
        self.coordinates = Coordinates(self.data["Lat"], self.data["Lon"])
        self.name = self.data["Name"]
        self.route_ids = self.data["Routes"]

    @property
    def routes(self) -> list["Route"]:
        """Return routes for this stop."""
        return [self.bus.routes[route_id] for route_id in self.route_ids]



"""NextBus models for MetroBus WMATA API."""
from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING, TypedDict

if TYPE_CHECKING:
    from .. import MetroBus
    from .route import Route


class NextBusData(TypedDict):
    """Next prediction data for MetroBus WMATA API."""

    DirectionNum: str
    DirectionText: str
    Minutes: int
    RouteID: str
    TripID: str
    VehicleID: str


@dataclass
class NextBus:
    """Next bus for a Stop."""

    bus: "MetroBus"
    data: NextBusData
    direction_number: int = field(init=False)
    direction_text: str = field(init=False)
    minutes: int = field(init=False)
    route_id: str = field(init=False)
    route: "Route" = field(init=False)
    trip_id: str = field(init=False)
    vehicle_id: str = field(init=False)

    def __post_init__(self) -> None:
        """Post init."""
        self.direction_number = int(self.data["DirectionNum"])
        self.direction_text = self.data["DirectionText"]
        self.minutes = self.data["Minutes"]
        self.route_id = self.data["RouteID"]
        self.route = self.bus.routes[self.route_id]
        self.trip_id = self.data["TripID"]
        self.vehicle_id = self.data["VehicleID"]



"""Route schedule models for MetroBus WMATA API."""
from __future__ import annotations

from dataclasses import dataclass, field
from datetime import datetime
from typing import TYPE_CHECKING, TypedDict

from ...const import TZ

if TYPE_CHECKING:
    from .. import MetroBus
    from .route import Route
    from .stop import Stop


class StopTimeData(TypedDict):
    """StopTime data for MetroBus WMATA API."""

    StopID: str
    StopName: str
    StopSeq: int
    Time: str


@dataclass
class StopTime:
    """Stop Time for a Route Schedule."""

    bus: "MetroBus"
    data: StopTimeData
    stop_id: str = field(init=False)
    stop_name: str = field(init=False)
    stop_sequence: int = field(init=False)
    time: datetime = field(init=False)

    def __post_init__(self) -> None:
        """Post init."""
        self.stop_id = self.data["StopID"]
        self.stop_name = self.data["StopName"]
        self.stop_sequence = self.data["StopSeq"]
        self.time = datetime.fromisoformat(self.data["Time"]).replace(tzinfo=TZ)

    @property
    def stop(self) -> Stop:
        """Get stop."""
        return self.bus.stops[self.stop_id]


class DirectionScheduleData(TypedDict):
    """DirectionSchedule data for MetroBus WMATA API."""

    DirectionNum: int  # deprecated
    EndTime: str
    RouteID: str
    StartTime: str
    StopTimes: list[StopTimeData]
    TripDirectionText: str
    TripHeadsign: str
    TripID: str


@dataclass
class DirectionSchedule:
    """Direction Schedule for a Path."""

    bus: "MetroBus"
    data: DirectionScheduleData
    direction_num: int
    end_time: datetime = field(init=False)
    route_id: str = field(init=False)
    route: Route = field(init=False)
    start_time: datetime = field(init=False)
    stop_times: list[StopTime] = field(init=False)
    trip_direction: str = field(init=False)
    trip_headsign: str = field(init=False)
    trip_id: str = field(init=False)

    def __post_init__(self) -> None:
        """Post init."""
        self.end_time = datetime.fromisoformat(self.data["EndTime"]).replace(tzinfo=TZ)
        self.route_id = self.data["RouteID"]
        self.route = self.bus.routes[self.route_id]
        self.start_time = datetime.fromisoformat(self.data["StartTime"]).replace(
            tzinfo=TZ
        )
        self.stop_times = sorted(
            [
                StopTime(self.bus, stop_time_data)
                for stop_time_data in self.data["StopTimes"]
            ],
            key=lambda stop_time: stop_time.stop_sequence,
        )
        self.trip_direction = self.data["TripDirectionText"]
        self.trip_headsign = self.data["TripHeadsign"]
        self.trip_id = self.data["TripID"]


class RouteScheduleData(TypedDict):
    """Route schedule data for MetroBus WMATA API."""

    Direction0: list[DirectionScheduleData] | None
    Direction1: list[DirectionScheduleData] | None


@dataclass
class RouteSchedule:
    """Schedule for a Route."""

    bus: "MetroBus"
    route: "Route"
    data: RouteScheduleData
    directions_schedules: dict[int, list[DirectionSchedule]] = field(
        init=False, default_factory=dict
    )

    def __post_init__(self) -> None:
        """Post init."""
        if (direction_schedules_data := self.data["Direction0"]) is not None:
            self.directions_schedules[0] = sorted(
                [
                    DirectionSchedule(self.bus, direction_schedule_data, 0)
                    for direction_schedule_data in direction_schedules_data
                ],
                key=lambda direction_schedule: direction_schedule.start_time,
            )
        if (direction_schedules_data := self.data["Direction1"]) is not None:
            self.directions_schedules[1] = sorted(
                [
                    DirectionSchedule(self.bus, direction_schedule_data, 1)
                    for direction_schedule_data in direction_schedules_data
                ],
                key=lambda direction_schedule: direction_schedule.start_time,
            )



"""Bus position models for MetroBus API."""
from __future__ import annotations

from dataclasses import dataclass, field
from datetime import datetime
from typing import TYPE_CHECKING, TypedDict

from ...const import TZ
from ...models.coordinates import Coordinates

if TYPE_CHECKING:
    from .. import MetroBus
    from .route import Route


class LiveBusPositionData(TypedDict):
    """Bus position data for MetroBus API."""

    DateTime: str
    Deviation: float
    DirectionNum: int  # deprecated
    DirectionText: str
    Lat: float
    Lon: float
    RouteID: str
    TripEndTime: str
    TripHeadsign: str
    TripID: str
    TripStartTime: str
    VehicleID: str
    BlockNumber: str


@dataclass
class LiveBusPosition:
    """A MetroBus Bus LiveBusPosition."""

    bus: "MetroBus"
    data: LiveBusPositionData
    last_update: datetime = field(init=False)
    deviation: float = field(init=False)
    direction_text: str = field(init=False)
    coordinates: Coordinates = field(init=False)
    route_id: str = field(init=False)
    trip_end_time: datetime = field(init=False)
    trip_headsign: str = field(init=False)
    trip_id: str = field(init=False)
    trip_start_time: datetime = field(init=False)
    vehicle_id: str = field(init=False)

    def __post_init__(self) -> None:
        """Post init."""
        self.last_update = datetime.fromisoformat(self.data["DateTime"]).replace(
            tzinfo=TZ
        )
        self.deviation = self.data["Deviation"]
        self.direction_text = self.data["DirectionText"]
        self.coordinates = Coordinates(self.data["Lat"], self.data["Lon"])
        self.route_id = self.data["RouteID"]
        self.trip_end_time = datetime.fromisoformat(self.data["TripEndTime"]).replace(
            tzinfo=TZ
        )
        self.trip_headsign = self.data["TripHeadsign"]
        self.trip_id = self.data["TripID"]
        self.trip_start_time = datetime.fromisoformat(
            self.data["TripStartTime"]
        ).replace(tzinfo=TZ)
        self.vehicle_id = self.data["VehicleID"]

    @property
    def route(self) -> "Route":
        """Get route."""
        return self.bus.routes[self.route_id]



"""Standard route models for MetroRail WMATA API."""
from __future__ import annotations

from dataclasses import dataclass, field
from typing import Literal, TypedDict


class TrackCircuitNeighborData(TypedDict):
    """Track circuit neighbor data for MetroRail WMATA API."""

    CircuitIds: list[int]
    NeighborType: Literal["Left", "Right"]


@dataclass
class TrackCircuitNeighbor:
    """Track circuit neighbor."""

    all_track_circuits: dict[int, TrackCircuit]
    data: TrackCircuitNeighborData
    circuit_ids: list[int] = field(init=False)
    neighbor_type: Literal["Left", "Right"] = field(init=False)

    def __post_init__(self) -> None:
        """Post init."""
        self.circuit_ids = self.data["CircuitIds"]
        self.neighbor_type = self.data["NeighborType"]

    @property
    def circuits(self) -> list[TrackCircuit]:
        """Circuits."""
        return [self.all_track_circuits[circuit_id] for circuit_id in self.circuit_ids]


class TrackCircuitData(TypedDict):
    """Track circuit data for MetroRail WMATA API."""

    CircuitId: int
    Neighbors: list[TrackCircuitNeighborData]
    Track: int


@dataclass
class TrackCircuit:
    """Track circuit."""

    all_track_circuits: dict[int, TrackCircuit]
    data: TrackCircuitData
    circuit_id: int = field(init=False)
    neighbors: list[TrackCircuitNeighbor] = field(init=False)
    track: int = field(init=False)

    def __post_init__(self) -> None:
        """Post init."""
        self.circuit_id = self.data["CircuitId"]
        self.neighbors = [
            TrackCircuitNeighbor(self.all_track_circuits, neighbor)
            for neighbor in self.data["Neighbors"]
        ]
        self.track = self.data["Track"]



"""Station parking information models for MetroRail WMATA API."""
from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING, TypedDict

if TYPE_CHECKING:
    from .. import MetroRail
    from .station import Station


class ShortTermParkingData(TypedDict):
    """Short-term parking data for MetroRail WMATA API."""

    TotalCount: int
    Notes: str | None


@dataclass
class ShortTermParking:
    """Short-term parking information for MetroRail WMATA API."""

    data: ShortTermParkingData
    total_count: int = field(init=False)
    notes: str | None = field(init=False)

    def __post_init__(self) -> None:
        """Post init."""
        self.total_count = self.data["TotalCount"]
        self.notes = self.data["Notes"] if self.data["Notes"] else None


class AllDayParkingData(TypedDict):
    """All-day parking data for MetroRail WMATA API."""

    TotalCount: int
    RiderCost: float | int
    NonRiderCost: float | int
    SaturdayRiderCost: float | int
    SaturdayNonRiderCost: float | int


@dataclass
class AllDayParking:
    """All-day parking information for MetroRail WMATA API."""

    data: AllDayParkingData
    total_count: int = field(init=False)
    rider_cost: float | None = field(init=False, default=None)
    non_rider_cost: float | None = field(init=False, default=None)
    saturday_rider_cost: float | None = field(init=False, default=None)
    saturday_non_rider_cost: float | None = field(init=False, default=None)

    def __post_init__(self) -> None:
        """Post init."""
        self.total_count = self.data["TotalCount"]

        if rider_cost := self.data["RiderCost"]:
            self.rider_cost = float(rider_cost)
        if non_rider_cost := self.data["NonRiderCost"]:
            self.non_rider_cost = float(non_rider_cost)
        if saturday_rider_cost := self.data["SaturdayRiderCost"]:
            self.saturday_rider_cost = float(saturday_rider_cost)
        if saturday_non_rider_cost := self.data["SaturdayNonRiderCost"]:
            self.saturday_non_rider_cost = float(saturday_non_rider_cost)


class StationParkingData(TypedDict):
    """Station parking data for MetroRail WMATA API."""

    Code: str
    Notes: str | None
    ShortTermParking: ShortTermParkingData
    AllDayParking: AllDayParkingData


@dataclass
class StationParking:
    """Station parking information for MetroRail WMATA API."""

    bus: "MetroRail"
    data: StationParkingData
    code: str = field(init=False)
    notes: str | None = field(init=False)
    short_term_parking: ShortTermParking = field(init=False)
    all_day_parking: AllDayParking = field(init=False)

    def __post_init__(self) -> None:
        """Post init."""
        self.code = self.data["Code"]
        self.notes = self.data["Notes"] if self.data["Notes"] else None
        self.short_term_parking = ShortTermParking(self.data["ShortTermParking"])
        self.all_day_parking = AllDayParking(self.data["AllDayParking"])

    @property
    def station(self) -> "Station":
        """Return the station."""
        return self.bus.stations[self.code]



"""Station models for MetroRail WMATA API."""
from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING, TypedDict

from ...models.coordinates import Coordinates
from .address import Address, AddressData

if TYPE_CHECKING:
    from .. import MetroRail
    from .line import Line
    from .station_entrance import StationEntrance
    from .station_parking import StationParking
    from .station_timings import StationTime


class StationData(TypedDict):
    """Station data for MetroRail WMATA API."""

    Address: "AddressData"
    Code: str
    Lat: float
    Lon: float
    LineCode1: str
    LineCode2: str | None
    LineCode3: str | None
    LineCode4: str | None
    Name: str
    StationTogether1: str
    StationTogether2: str


@dataclass
class Station:
    """MetroRail Station."""

    bus: "MetroRail"
    data: StationData
    parking: "StationParking" | None
    entrances: list["StationEntrance"]
    station_times: list["StationTime"]
    address: Address = field(init=False)
    station_code: str = field(init=False)
    coordinates: Coordinates = field(init=False)
    line_codes: list[str] = field(init=False)
    name: str = field(init=False)
    station_together_code_1: str | None = field(init=False)
    station_together_code_2: str | None = field(init=False)

    def __post_init__(self) -> None:
        """Post init."""
        self.address = Address(self.data["Address"])
        self.station_code = self.data["Code"]
        self.coordinates = Coordinates(self.data["Lat"], self.data["Lon"])
        self.line_codes = [
            self.data[f"LineCode{x}"]  # type: ignore
            for x in range(1, 5)
            if self.data[f"LineCode{x}"]  # type: ignore
        ]
        self.name = self.data["Name"]
        self.station_together_code_1 = (
            self.data["StationTogether1"] if self.data["StationTogether1"] else None
        )
        self.station_together_code_2 = (
            self.data["StationTogether2"] if self.data["StationTogether2"] else None
        )

    @property
    def lines(self) -> list["Line"]:
        """Return the lines."""
        return [self.bus.lines[line_code] for line_code in self.line_codes]

    @property
    def station_together_1(self) -> "Station" | None:
        """Return the station together 1."""
        if not self.station_together_code_1:
            return None
        return self.bus.stations[self.station_together_code_1]

    @property
    def station_together_2(self) -> "Station" | None:
        """Return the station together 2."""
        if not self.station_together_code_2:
            return None
        return self.bus.stations[self.station_together_code_2]



"""Bus incident models for MetroBus WMATA API."""
from __future__ import annotations

from dataclasses import dataclass, field
from datetime import datetime
from typing import TYPE_CHECKING, TypedDict

from ...const import TZ

if TYPE_CHECKING:
    from .. import MetroBus
    from .route import Route


class BusIncidentData(TypedDict):
    """Incident data for MetroBus WMATA API."""

    DateUpdated: str
    Description: str
    IncidentID: str
    IncidentType: str
    RoutesAffected: list[str]


@dataclass
class BusIncident:
    """MetroBus Incident."""

    bus: "MetroBus"
    data: BusIncidentData
    date_updated: datetime = field(init=False)
    description: str = field(init=False)
    incident_id: str = field(init=False)
    incident_type: str = field(init=False)
    route_ids_affected: list[str] = field(init=False)

    def __post_init__(self) -> None:
        """Post init."""
        self.date_updated = datetime.fromisoformat(self.data["DateUpdated"]).replace(
            tzinfo=TZ
        )
        self.description = self.data["Description"]
        self.incident_type = self.data["IncidentType"]
        self.incident_id = self.data["IncidentID"]
        self.route_ids_affected = self.data["RoutesAffected"]

    @property
    def routes_affected(self) -> list["Route"]:
        """Return routes affected by this incident."""
        return [self.bus.routes[route_id] for route_id in self.route_ids_affected]



"""Station entrance models for MetroRail WMATA API."""
from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING, TypedDict

from ...models.coordinates import Coordinates

if TYPE_CHECKING:
    from .. import MetroRail
    from .station import Station


class StationEntranceData(TypedDict):
    """Station entrance data for MetroRail WMATA API."""

    Description: str
    ID: str
    Lat: float
    Lon: float
    Name: str
    StationCode1: str
    StationCode2: str


@dataclass
class StationEntrance:
    """Station entrance information for MetroRail WMATA API."""

    bus: "MetroRail"
    data: StationEntranceData
    description: str = field(init=False)
    entrance_id: str = field(init=False)
    coordinates: Coordinates = field(init=False)
    name: str = field(init=False)
    station_code_1: str = field(init=False)
    station_code_2: str = field(init=False)

    def __post_init__(self) -> None:
        """Post init."""
        self.description = self.data["Description"]
        self.entrance_id = self.data["ID"]
        self.coordinates = Coordinates(self.data["Lat"], self.data["Lon"])
        self.name = self.data["Name"]
        self.station_code_1 = self.data["StationCode1"]
        self.station_code_2 = self.data["StationCode2"]

    @property
    def station_1(self) -> "Station":
        """Return the station 1."""
        return self.bus.stations[self.station_code_1]

    @property
    def station_2(self) -> "Station":
        """Return the station 2."""
        return self.bus.stations[self.station_code_2]



"""Module for interacting with MetroRail API."""
from __future__ import annotations

import asyncio
from collections import defaultdict
from typing import TYPE_CHECKING, cast

from .const import RailEndpoint
from .models.elevator_and_escalator_incident import ElevatorAndEscalatorIncident
from .models.line import Line
from .models.live_position import LiveTrainPosition
from .models.next_train import NextTrain
from .models.rail_incident import RailIncident
from .models.standard_route import StandardRoute
from .models.station import Station
from .models.station_entrance import StationEntrance
from .models.station_parking import StationParking
from .models.station_timings import StationTime
from .models.station_to_station import StationToStation, StationToStationPathData
from .models.track_circuit import TrackCircuit

if TYPE_CHECKING:
    from ..client import Client


class MetroRail:
    """Class to represent client for MetroRail APIs."""

    client: "Client"
    lines: dict[str, Line]
    stations: dict[str, Station]

    def __init__(self, client: "Client") -> None:
        """Initialize."""
        self.client = client
        self.lines = {}
        self.stations = {}

    async def load_data(self) -> None:
        """Load the base data."""
        self.lines = await self.get_all_lines()
        self.stations = await self.get_stations()

    async def get_all_lines(self) -> dict[str, Line]:
        """Get all lines."""
        standard_routes_data, lines_data = await asyncio.gather(
            self.client.fetch(
                RailEndpoint.STANDARD_ROUTES, params={"contentType": "json"}
            ),
            self.client.fetch(RailEndpoint.LINES),
        )
        standard_routes = defaultdict(list)
        for standard_route_data in standard_routes_data["StandardRoutes"]:
            standard_route = StandardRoute(self, standard_route_data)
            standard_routes[standard_route.line_code].append(standard_route)
        return {
            line["LineCode"]: Line(self, line, standard_routes[line["LineCode"]])
            for line in lines_data["Lines"]
        }

    def __get_entrances(
        self, entrances_data: dict
    ) -> defaultdict[str, list[StationEntrance]]:
        """Get entrances from entrances data."""
        entrances = defaultdict(list)
        for entrance_data in entrances_data["Entrances"]:
            entrance = StationEntrance(self, entrance_data)
            for code_num in range(1, 3):
                if code := entrance_data[f"StationCode{code_num}"]:
                    entrances[code].append(entrance)
        return entrances

    def __get_station_times(
        self, stations_times_data: dict
    ) -> defaultdict[str, list[StationTime]]:
        """Get station times from station times data."""
        stations_times = defaultdict(list)
        for station_time_data in stations_times_data["StationTimes"]:
            station_time = StationTime(self, station_time_data)
            stations_times[station_time.station_code].append(station_time)
        return stations_times

    async def get_stations(self, line: Line | None = None) -> dict[str, Station]:
        """Get all stations."""
        params = {}
        if line:
            params = {"LineCode": line.line_code}
        (
            station_parking_data,
            entrances_data,
            stations_times_data,
            stations_data,
        ) = await asyncio.gather(
            self.client.fetch(RailEndpoint.STATION_PARKING_INFORMATION),
            self.client.fetch(RailEndpoint.STATION_ENTRANCES),
            self.client.fetch(RailEndpoint.STATION_TIMINGS),
            self.client.fetch(RailEndpoint.STATIONS, params=params),
        )

        station_parking = {
            station_parking["Code"]: StationParking(self, station_parking)
            for station_parking in station_parking_data["StationsParking"]
        }
        entrances = self.__get_entrances(entrances_data)
        stations_times = self.__get_station_times(stations_times_data)

        return {
            station_data["Code"]: Station(
                self,
                station_data,
                station_parking.get(station_data["Code"]),
                entrances[station_data["Code"]],
                stations_times[station_data["Code"]],
            )
            for station_data in stations_data["Stations"]
        }

    async def get_station_to_station_data(
        self, from_station: Station, to_station: Station
    ) -> StationToStation:
        """Get the path from one station to another."""
        params = {
            "FromStationCode": from_station.station_code,
            "ToStationCode": to_station.station_code,
        }
        path_data = cast(
            StationToStationPathData,
            await self.client.fetch(
                RailEndpoint.STATION_TO_STATION_PATH, params=params
            ),
        )
        info_data = await self.client.fetch(
            RailEndpoint.STATION_TO_STATION_INFO, params=params
        )
        return StationToStation(self, path_data, info_data["StationToStationInfos"])

    async def get_elevator_escalator_incidents(
        self, station: Station | None = None
    ) -> list[ElevatorAndEscalatorIncident]:
        """Get elevator and escalator incidents."""
        params = {}
        if station:
            params["StationCode"] = station.station_code
        data = await self.client.fetch(
            RailEndpoint.ELEVATOR_ESCALATOR_INCIDENTS, params=params
        )
        return sorted(
            [
                ElevatorAndEscalatorIncident(self, elevator_escalator_data)
                for elevator_escalator_data in data["ElevatorIncidents"]
            ],
            reverse=True,
            key=lambda incident: incident.date_updated,
        )

    async def get_rail_incidents(self) -> list[RailIncident]:
        """Get rail incidents."""
        data = await self.client.fetch(RailEndpoint.RAIL_INCIDENTS)
        return sorted(
            [RailIncident(self, incident_data) for incident_data in data["Incidents"]],
            reverse=True,
            key=lambda incident: incident.date_updated,
        )

    async def get_next_trains_at_station(
        self, stations: Station | list[Station]
    ) -> list[NextTrain]:
        """Return next trains for given station(s)."""

        def key_func(train: NextTrain) -> int:
            if not train.minutes:
                return -3
            if train.minutes == "BRD":
                return -2
            if train.minutes == "ARR":
                return -1
            assert isinstance(train.minutes, int)
            return train.minutes

        if not isinstance(stations, list):
            stations = [stations]
        station_codes = ",".join([station.station_code for station in stations])
        data = await self.client.fetch(
            RailEndpoint.NEXT_TRAINS, additional_path=station_codes
        )
        return sorted(
            [NextTrain(self, next_train_data) for next_train_data in data["Trains"]],
            key=key_func,
        )

    async def get_live_positions(self) -> dict[str, LiveTrainPosition]:
        """Get live train positions."""
        data = await self.client.fetch(
            RailEndpoint.TRAIN_POSITIONS, params={"contentType": "json"}
        )
        return {
            train_position["TrainId"]: LiveTrainPosition(self, train_position)
            for train_position in data["TrainPositions"]
        }

    async def get_track_circuits(self) -> dict[int, TrackCircuit]:
        """Get track circuits."""
        track_circuits_data = await self.client.fetch(
            RailEndpoint.TRACK_CIRCUITS, params={"contentType": "json"}
        )
        track_circuits: dict[int, TrackCircuit] = {}
        for track_circuit in track_circuits_data["TrackCircuits"]:
            track_circuits[track_circuit["CircuitId"]] = TrackCircuit(
                track_circuits, track_circuit
            )
        return track_circuits



"""Constants for MetroBus API."""
from __future__ import annotations

from enum import Enum

from ..const import BASE_URL


class BusEndpoint(Enum):
    """Endpoints for all MetroBus APIs."""

    BUS_INCIDENTS = f"{BASE_URL}/Incidents.svc/json/BusIncidents"
    NEXT_BUSES = f"{BASE_URL}/NextBusService.svc/json/jNextBuss"
    ROUTE_PATH = f"{BASE_URL}/Bus.svc/json/jRouteDetails"
    POSITIONS = f"{BASE_URL}/Bus.svc/json/jBusLiveBusPositions"
    ROUTE_SCHEDULE = f"{BASE_URL}/Bus.svc/json/jRouteSchedule"
    ROUTES = f"{BASE_URL}/Bus.svc/json/jRoutes"
    STOP_SCHEDULE = f"{BASE_URL}/Bus.svc/json/jStopSchedule"
    STOPS = f"{BASE_URL}/Bus.svc/json/jStops"



"""Station to station path models for MetroRail WMATA API."""
from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING, TypedDict

if TYPE_CHECKING:
    from .. import MetroRail
    from .line import Line
    from .station import Station


class PathItemData(TypedDict):
    """Path item data for MetroRail WMATA API."""

    DistanceToPrev: int
    LineCode: str
    SeqNum: int
    StationCode: str
    StationName: str


@dataclass
class PathItem:
    """MetroRail Path Item."""

    bus: "MetroRail"
    data: PathItemData
    distance_to_previous: int = field(init=False)
    line_code: str = field(init=False)
    sequence_number: int = field(init=False)
    station_code: str = field(init=False)
    station_name: str = field(init=False)

    def __post_init__(self) -> None:
        """Post init."""
        self.distance_to_previous = self.data["DistanceToPrev"]
        self.line_code = self.data["LineCode"]
        self.sequence_number = self.data["SeqNum"]
        self.station_code = self.data["StationCode"]
        self.station_name = self.data["StationName"]

    @property
    def line(self) -> "Line":
        """Return the line."""
        return self.bus.lines[self.line_code]

    @property
    def station(self) -> "Station":
        """Return the station."""
        return self.bus.stations[self.station_code]


class StationToStationPathData(TypedDict):
    """Station to station path data for MetroRail WMATA API."""

    Path: list[PathItemData]


class RailFareData(TypedDict):
    """Rail fare data for MetroRail WMATA API."""

    OffPeakTime: float | int
    PeakTime: float | int
    SeniorDisabled: float | int


@dataclass
class RailFare:
    """MetroRail Rail Fare."""

    data: RailFareData
    off_peak_time: float = field(init=False)
    peak_time: float = field(init=False)
    senior_disabled: float = field(init=False)

    def __post_init__(self) -> None:
        """Post init."""
        self.off_peak_time = float(self.data["OffPeakTime"])
        self.peak_time = float(self.data["PeakTime"])
        self.senior_disabled = float(self.data["SeniorDisabled"])


class StationToStationInfoData(TypedDict):
    """Station to station info data for MetroRail WMATA API."""

    CompositeMiles: float | int
    DestinationStation: str
    RailFare: RailFareData
    RailTime: str
    SourceStation: str


@dataclass
class StationToStationInformation:
    """MetroRail Station to Station Info."""

    bus: "MetroRail"
    data: StationToStationInfoData
    composite_miles: float = field(init=False)
    destination_station_code: str = field(init=False)
    rail_fare: RailFare = field(init=False)
    rail_time: str = field(init=False)
    source_station_code: str = field(init=False)

    def __post_init__(self) -> None:
        """Post init."""
        self.composite_miles = float(self.data["CompositeMiles"])
        self.destination_station_code = self.data["DestinationStation"]
        self.rail_fare = RailFare(self.data["RailFare"])
        self.rail_time = self.data["RailTime"]
        self.source_station_code = self.data["SourceStation"]

    @property
    def destination_station(self) -> "Station":
        """Return the destination station."""
        return self.bus.stations[self.destination_station_code]

    @property
    def source_station(self) -> "Station":
        """Return the source station."""
        return self.bus.stations[self.source_station_code]


class StationToStationData(TypedDict):
    """Station to station data for MetroRail WMATA API."""

    path: StationToStationPathData
    info: StationToStationInfoData


@dataclass
class StationToStation:
    """MetroRail Station to Station."""

    bus: "MetroRail"
    path_data: StationToStationPathData
    info_data: list[StationToStationInfoData]
    path: list[PathItem] = field(init=False)
    info: list[StationToStationInformation] = field(init=False)

    def __post_init__(self) -> None:
        """Post init."""
        self.path = sorted(
            [PathItem(self.bus, item_data) for item_data in self.path_data["Path"]],
            key=lambda item: item.sequence_number,
        )
        self.info = [
            StationToStationInformation(self.bus, station_info_data)
            for station_info_data in self.info_data
        ]



"""Address models for MetroRail WMATA API."""
from __future__ import annotations

from dataclasses import dataclass, field
from typing import TypedDict


class AddressData(TypedDict):
    """Station address data for MetroRail WMATA API."""

    City: str
    State: str
    Street: str
    Zip: str


@dataclass
class Address:
    """MetroRail Station Address."""

    data: AddressData
    street: str = field(init=False)
    city: str = field(init=False)
    state: str = field(init=False)
    zip_code: str = field(init=False)

    def __post_init__(self) -> None:
        """Post init."""
        self.street = self.data["Street"]
        self.city = self.data["City"]
        self.state = self.data["State"]
        self.zip_code = self.data["Zip"]



"""Test pywmataio client for trains."""
from datetime import date, datetime, time

from wmataio.client import Client
from wmataio.const import TZ
from wmataio.rail.models.live_position import TrainDirection


async def test_rail_apis(aioresponses):
    """Callback for aioresponse."""
    client = Client("", test_mode=True)
    await client.rail.load_data()
    assert client.rail.lines
    assert client.rail.stations

    assert len(client.rail.lines) == 6
    line = client.rail.lines["BL"]
    assert line.line_code == "BL"
    assert line.display_name == "Blue"
    assert line.start_station_code == "J03"
    assert line.start_station == client.rail.stations["J03"]
    assert line.end_station_code == "G05"
    assert line.end_station == client.rail.stations["G05"]
    assert line.internal_destination_code_1 is None
    assert line.internal_destination_1 is None
    assert line.internal_destination_code_2 is None
    assert line.internal_destination_2 is None

    assert len(line.standard_routes) == 2
    standard_route = line.standard_routes[0]
    assert standard_route.line_code == "BL"
    assert standard_route.line == line
    assert standard_route.track_number == 1
    assert len(standard_route.track_circuits) == 415

    track_circuit = standard_route.track_circuits[0]
    assert track_circuit.sequence_number == 0
    assert track_circuit.circuit_id == 2603
    assert track_circuit.station_code is None
    assert track_circuit.station is None

    assert len(client.rail.stations) == 101
    station = client.rail.stations["A01"]
    assert station.station_code == "A01"
    assert station.name == "Metro Center"
    assert station.station_together_code_1 == "C01"
    assert station.station_together_1 == client.rail.stations["C01"]
    assert station.station_together_code_2 is None
    assert station.station_together_2 is None
    assert station.line_codes == ["RD"]
    assert station.lines == [client.rail.lines["RD"]]
    assert station.coordinates.latitude == 38.898303
    assert station.coordinates.longitude == -77.028099
    assert station.address.street == "607 13th St. NW"
    assert station.address.city == "Washington"
    assert station.address.state == "DC"
    assert station.address.zip_code == "20005"
    assert len(station.entrances) == 8
    assert len(station.station_times) == 1
    assert station.parking is None

    entrance = station.entrances[0]
    assert entrance.entrance_id == "273"
    assert entrance.name == "11TH ST NW & G ST NW"
    assert entrance.station_code_1 == "A01"
    assert entrance.station_1 == client.rail.stations["A01"]
    assert entrance.station_code_2 == "C01"
    assert entrance.station_2 == client.rail.stations["C01"]
    assert entrance.description == (
        "Building Entrance from the southeast corner of 11th St NW and G St NW."
    )
    assert entrance.coordinates.latitude == 38.898073
    assert entrance.coordinates.longitude == -77.026789

    station_time = station.station_times[0]
    assert station_time.station_code == "A01"
    assert station_time.station == client.rail.stations["A01"]
    assert station_time.name == "Metro Center"

    day_time = station_time.get_day_time(date(2023, 3, 31))
    assert day_time.day == "Friday"
    assert day_time.day_of_week == 5
    assert day_time.opening_time == time(5, 14, 0, tzinfo=TZ)
    assert len(day_time.first_trains) == 2
    assert len(day_time.last_trains) == 2

    train_timing = day_time.first_trains[0]
    assert train_timing.time == time(5, 31, 0, tzinfo=TZ)
    assert train_timing.destination_station_code == "A15"
    assert train_timing.destination_station == client.rail.stations["A15"]

    parking = client.rail.stations["A07"].parking
    assert parking.code == "A07"
    assert parking.notes is None
    assert parking.all_day_parking.total_count == 0
    assert parking.all_day_parking.rider_cost is None
    assert parking.all_day_parking.non_rider_cost is None
    assert parking.all_day_parking.saturday_rider_cost is None
    assert parking.all_day_parking.saturday_non_rider_cost is None
    assert parking.short_term_parking.total_count == 17
    assert parking.short_term_parking.notes == (
        "Parking available 8:30 AM - 3:30 PM and 7 PM - 2 AM."
    )

    station_to_station = await client.rail.get_station_to_station_data(
        client.rail.stations["A15"], client.rail.stations["A01"]
    )
    assert len(station_to_station.path) == 15
    path_item = station_to_station.path[0]
    assert path_item.line_code == "RD"
    assert path_item.line == client.rail.lines["RD"]
    assert path_item.station_code == "A15"
    assert path_item.station == client.rail.stations["A15"]
    assert path_item.sequence_number == 1
    assert path_item.distance_to_previous == 0
    assert len(station_to_station.info) == 1
    info = station_to_station.info[0]
    assert info.source_station_code == "A15"
    assert info.source_station == client.rail.stations["A15"]
    assert info.destination_station_code == "A01"
    assert info.destination_station == client.rail.stations["A01"]
    assert info.composite_miles == 17.44
    assert info.rail_time == 36
    assert info.rail_fare.peak_time == 6.0
    assert info.rail_fare.off_peak_time == 3.85
    assert info.rail_fare.senior_disabled == 3.0

    station_to_station = await client.rail.get_station_to_station_data(
        client.rail.stations["A15"], client.rail.stations["N12"]
    )
    assert len(station_to_station.path) == 0

    elevator_escalator_incidents = await client.rail.get_elevator_escalator_incidents(
        client.rail.stations["A01"]
    )
    assert len(elevator_escalator_incidents) == 1

    elevator_escalator_incidents = await client.rail.get_elevator_escalator_incidents()
    assert len(elevator_escalator_incidents) == 34
    elevator_escalator_incident = elevator_escalator_incidents[0]
    assert elevator_escalator_incident.station_code == "C05"
    assert elevator_escalator_incident.station == client.rail.stations["C05"]
    assert elevator_escalator_incident.station_name == "Rosslyn"
    assert elevator_escalator_incident.unit_type == "ELEVATOR"
    assert elevator_escalator_incident.unit_name == "C05E03"
    assert elevator_escalator_incident.location_description == (
        "Elevator between street, and upper platform"
    )
    assert elevator_escalator_incident.symptom_description == "Service Call"
    assert elevator_escalator_incident.date_out_of_service == datetime(
        2023, 3, 31, 21, 5, 0, tzinfo=TZ
    )
    assert elevator_escalator_incident.date_updated == datetime(
        2023, 3, 31, 22, 11, 57, tzinfo=TZ
    )
    assert elevator_escalator_incident.estimated_return_to_service == datetime(
        2023, 4, 2, 23, 59, 59, tzinfo=TZ
    )

    rail_incidents = await client.rail.get_rail_incidents()
    assert len(rail_incidents) == 1
    rail_incident = rail_incidents[0]
    assert rail_incident.incident_id == "038F7C95-840B-4830-8EC3-ACAA9390DA09"
    assert rail_incident.description == (
        "No YL train service due to the bridge & tunnel project until May 2023. "
        "Use shuttle buses or BL/GR Line trains as alternate travel options."
    )
    assert rail_incident.incident_type == "Alert"
    assert rail_incident.line_codes_affected == ["YL"]
    assert rail_incident.lines_affected == [client.rail.lines["YL"]]
    assert rail_incident.date_updated == datetime(2023, 3, 8, 5, 23, 33, tzinfo=TZ)

    next_trains = await client.rail.get_next_trains_at_station(
        client.rail.stations["A15"]
    )
    assert len(next_trains) == 2
    next_train = next_trains[0]
    assert next_train.num_cars is None
    assert next_train.destination_station_code == "B11"
    assert next_train.destination_station == client.rail.stations["B11"]
    assert next_train.destination_station_name == "Glenmont"
    assert next_train.group == 2
    assert next_train.line_code == "RD"
    assert next_train.line == client.rail.lines["RD"]
    assert next_train.location_code == "A15"
    assert next_train.location == client.rail.stations["A15"]
    assert next_train.location_name == "Shady Grove"
    assert next_train.minutes == 9

    live_positions = await client.rail.get_live_positions()
    assert len(live_positions) == 82
    live_position = list(live_positions.values())[0]
    assert live_position.train_id == "075"
    assert live_position.train_number == "000"
    assert live_position.car_count == 6
    assert live_position.direction == TrainDirection.SOUTHBOUND_OR_WESTBOUND
    assert live_position.circuit_id == 1
    assert live_position.destination_station_code is None
    assert live_position.destination_station is None
    assert live_position.line_code is None
    assert live_position.line is None
    assert live_position.seconds_at_location == 12157
    assert live_position.service_type == "Unknown"

    track_circuits = await client.rail.get_track_circuits()
    assert len(track_circuits) == 3315
    track_circuit = list(track_circuits.values())[0]
    assert track_circuit.track == 1
    assert track_circuit.circuit_id == 1
    assert len(track_circuit.neighbors) == 1
    neighbor = track_circuit.neighbors[0]
    assert neighbor.neighbor_type == "Right"
    assert neighbor.circuit_ids == [2]
    assert neighbor.circuits == [track_circuits[2]]



"""Train incident models for MetroRail WMATA API."""
from __future__ import annotations

from dataclasses import dataclass, field
from datetime import datetime
from typing import TYPE_CHECKING, TypedDict

from ...const import TZ

if TYPE_CHECKING:
    from .. import MetroRail
    from .line import Line


class RailIncidentData(TypedDict):
    """Incident data for MetroRail WMATA API."""

    DateUpdated: str
    DelaySeverity: str | None  # deprecated
    Description: str
    EmergencyText: str | None  # deprecated
    EndLocationFullName: str | None  # deprecated
    IncidentID: str
    IncidentType: str
    LinesAffected: str
    PassengerDelay: int | None  # deprecated
    StartLocationFullName: str | None  # deprecated


@dataclass
class RailIncident:
    """MetroRail Incident."""

    bus: "MetroRail"
    data: RailIncidentData
    date_updated: datetime = field(init=False)
    description: str = field(init=False)
    incident_id: str = field(init=False)
    incident_type: str = field(init=False)
    line_codes_affected: list[str] = field(init=False)

    def __post_init__(self) -> None:
        """Post init."""
        self.date_updated = datetime.fromisoformat(self.data["DateUpdated"]).replace(
            tzinfo=TZ
        )
        self.description = self.data["Description"]
        self.incident_id = self.data["IncidentID"]
        self.incident_type = self.data["IncidentType"]
        self.line_codes_affected = [
            line_code.strip()
            for line_code in self.data["LinesAffected"].split(";")
            if line_code.strip()
        ]

    @property
    def lines_affected(self) -> list["Line"]:
        """Lines affected."""
        return [self.bus.lines[line_code] for line_code in self.line_codes_affected]



"""Base Client logic for WMATA API clients."""
from __future__ import annotations

import asyncio
import logging
from contextlib import nullcontext
from dataclasses import dataclass, field
from json.decoder import JSONDecodeError
from typing import Any, cast

from aiohttp import ClientSession, client_exceptions

from .bus import MetroBus
from .bus.const import BusEndpoint
from .const import ADDITIONAL_PATH_HEADER, CLASS_HEADER, ENUM_HEADER
from .exceptions import WMATAError
from .rail import MetroRail
from .rail.const import RailEndpoint

_LOGGER = logging.getLogger(__name__)


@dataclass
class Client:
    """Client to provide API request support."""

    api_key: str
    session: ClientSession | None = None
    test_mode: bool = False
    bus: MetroBus = field(init=False)
    rail: MetroRail = field(init=False)
    _headers: dict[str, str] = field(init=False, default_factory=dict)

    def __post_init__(self) -> None:
        """Post initialize."""
        self.bus = MetroBus(self)
        self.rail = MetroRail(self)
        self._headers = {"api_key": self.api_key}

    def _get_headers(
        self, enum_: BusEndpoint | RailEndpoint, additional_path: str | None = None
    ) -> dict[str, Any]:
        """Get headers."""
        headers = self._headers
        if self.test_mode:
            return {
                **headers,
                CLASS_HEADER: enum_.__class__.__name__,
                ENUM_HEADER: enum_.name,
                ADDITIONAL_PATH_HEADER: additional_path or "",
            }
        return headers

    async def fetch(
        self,
        enum_: BusEndpoint | RailEndpoint,
        params: dict[str, Any] | None = None,
        additional_path: str | None = None,
    ) -> dict:
        """Fetch data from WMATA API."""
        url = enum_.value
        if additional_path:
            url = f"{url}/{additional_path}"

        _LOGGER.debug("Fetching %s with params %s", url, params or {})
        retry = True
        context_manager = nullcontext(self.session) if self.session else ClientSession()
        async with context_manager as session:
            while retry:
                try:
                    response = await session.get(
                        url,
                        params=params,
                        headers=self._get_headers(enum_, additional_path),
                        raise_for_status=True,
                    )
                except client_exceptions.ClientResponseError as error:
                    if error.status != 429:
                        raise WMATAError("Error while making request") from error
                    _LOGGER.warning("Too many requests, sleeping for 1.1 seconds")
                    await asyncio.sleep(1.1)
                except client_exceptions.ClientError as error:
                    raise WMATAError("Error while making request") from error
                else:
                    retry = False

            try:
                response_json = await response.json(content_type=None)
            except JSONDecodeError as error:
                _LOGGER.error("Invalid JSON: %s", response.text())
                raise WMATAError("Invalid JSON") from error
            _LOGGER.debug("Response: %s", response_json)

            try:
                return cast(dict, response_json)
            except Exception as error:
                raise WMATAError("Could not parse response json into object") from error



"""Fixtures for pywmataio tests."""
import json
import logging
import pathlib
import re

import pytest
from aioresponses import CallbackResult, aioresponses

from wmataio.bus.const import BusEndpoint
from wmataio.const import ADDITIONAL_PATH_HEADER, CLASS_HEADER, ENUM_HEADER
from wmataio.rail.const import RailEndpoint

CLASS_MAP = {
    BusEndpoint.__name__: {"class": BusEndpoint, "api_type": "bus"},
    RailEndpoint.__name__: {"class": RailEndpoint, "api_type": "rail"},
}


_LOGGER = logging.getLogger(__name__)
API_REGEX = re.compile(r"https:\/\/api\.wmata\.com.*$")


def _callback(url: str, **kwargs) -> CallbackResult:
    """Respond to WMATA API calls."""
    _LOGGER.debug("Received request for %s", url)
    params = kwargs.get("params")
    headers = kwargs["headers"]
    class_name = headers[CLASS_HEADER]
    enum_name = headers[ENUM_HEADER]
    additional_path = headers[ADDITIONAL_PATH_HEADER]
    cls_data = CLASS_MAP[class_name]
    enum_: BusEndpoint | RailEndpoint = cls_data["class"][enum_name]
    api_type = cls_data["api_type"]

    base_fixture_name = enum_.name.lower()

    additional_path_fixture_name = ""
    if additional_path:
        additional_path_fixture_name = additional_path.replace("/", "_")

    params_fixture_name = ""
    if params:
        params_fixture_name = "_".join(f"{k}_{v}" for k, v in params.items())

    fixture_name = base_fixture_name
    if additional_path_fixture_name or params_fixture_name:
        fixture_name = ".".join(
            [base_fixture_name, additional_path_fixture_name, params_fixture_name]
        )

    path = pathlib.Path(f"test/fixtures/{api_type}/{fixture_name}.json")

    if not path.exists():
        _LOGGER.error("Fixture %s does not exist", path)
        return CallbackResult(status=404)

    with open(path, "r") as fp:
        data = json.load(fp)

    return CallbackResult(status=200, payload=data)


@pytest.fixture(name="aioresponses", scope="session")
def mock_aioresponse():
    """Mock aiohttp responses."""
    with aioresponses() as m:
        m.get(API_REGEX, callback=_callback, repeat=True)
        yield m



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pywmataio",
    version="0.0.1",
    author="Raman",
    author_email="7243222+raman325@users.noreply.github.com",
    description="A simple async interface to the Washington, DC WMATA API",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/raman325/pywmataio",
    packages=setuptools.find_packages(),
    install_requires=["aiohttp"],
    tests_require=[
        "aioresponses",
        "pytest",
        "pytest-asyncio",
        "pytest-cov",
        "pytest-timeout",
    ],
    classifiers=[
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
    ],
    python_requires=">=3.10",
)



"""Stop schedule models for MetroBus WMATA API."""
from __future__ import annotations

from dataclasses import dataclass, field
from datetime import datetime
from typing import TYPE_CHECKING, TypedDict

from ...const import TZ

if TYPE_CHECKING:
    from .. import MetroBus
    from .route import Route
    from .stop import Stop


class StopArrivalData(TypedDict):
    """ArrivalSchedule data for MetroBus WMATA API."""

    DirectionNum: str
    EndTime: str
    RouteID: str
    ScheduleTime: str
    StartTime: str
    TripDirectionText: str
    TripHeadsign: str
    TripID: str


@dataclass
class StopArrival:
    """Stop Arrival for a Route Schedule."""

    bus: "MetroBus"
    stop: "Stop"
    data: StopArrivalData
    direction_number: int = field(init=False)
    end_time: datetime = field(init=False)
    route_id: str = field(init=False)
    route: "Route" = field(init=False)
    schedule_time: datetime = field(init=False)
    start_time: datetime = field(init=False)
    trip_direction_text: str = field(init=False)
    trip_headsign: str = field(init=False)
    trip_id: str = field(init=False)

    def __post_init__(self) -> None:
        """Post init."""
        self.direction_number = int(self.data["DirectionNum"])
        self.end_time = datetime.fromisoformat(self.data["EndTime"]).replace(tzinfo=TZ)
        self.route_id = self.data["RouteID"]
        self.route = self.bus.routes[self.route_id]
        self.schedule_time = datetime.fromisoformat(self.data["ScheduleTime"]).replace(
            tzinfo=TZ
        )
        self.start_time = datetime.fromisoformat(self.data["StartTime"]).replace(
            tzinfo=TZ
        )
        self.trip_direction_text = self.data["TripDirectionText"]
        self.trip_headsign = self.data["TripHeadsign"]
        self.trip_id = self.data["TripID"]



"""Path models for MetroBus WMATA API."""
from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING, TypedDict

from ...models.coordinates import Coordinates

if TYPE_CHECKING:
    from .. import MetroBus
    from .route import Route
    from .stop import Stop, StopData


class ShapePointData(TypedDict):
    """ShapePoint data for MetroBus WMATA API."""

    Lat: float | int
    Lon: float | int
    SeqNum: str


@dataclass
class ShapePoint:
    """ShapePoint for a Path."""

    data: ShapePointData
    coordinates: Coordinates = field(init=False)
    sequence_number: int = field(init=False)

    def __post_init__(self) -> None:
        """Post init."""
        self.coordinates = Coordinates(float(self.data["Lat"]), float(self.data["Lon"]))
        self.sequence_number = int(self.data["SeqNum"])


class PathDirectionData(TypedDict):
    """Direction data for MetroBus WMATA API."""

    DirectionNum: int  # deprecated
    DirectionText: str
    Shape: list[ShapePointData]
    Stops: list["StopData"]
    TripHeadsign: str


@dataclass
class PathDirection:
    """Direction for a Path."""

    bus: "MetroBus"
    data: PathDirectionData
    direction_num: int
    direction: str = field(init=False)
    shapes: list[ShapePoint] = field(init=False)
    stops: list["Stop"] = field(init=False)
    trip_headsign: str = field(init=False)

    def __post_init__(self) -> None:
        """Post init."""
        self.direction = self.data["DirectionText"]
        self.shapes = sorted(
            [ShapePoint(shape_data) for shape_data in self.data["Shape"]],
            key=lambda shape: shape.sequence_number,
        )
        self.stops = [
            self.bus.get_stop_from_stop_data(stop_data)
            for stop_data in self.data["Stops"]
        ]
        self.trip_headsign = self.data["TripHeadsign"]


class RoutePathData(TypedDict):
    """Path data for MetroBus WMATA API."""

    RouteID: str
    Name: str
    Direction0: PathDirectionData | None
    Direction1: PathDirectionData | None


@dataclass
class RoutePath:
    """Path for a MetroBus."""

    bus: "MetroBus"
    data: RoutePathData
    route_id: str = field(init=False)
    route: "Route" = field(init=False)
    name: str = field(init=False)
    directions: dict[int, PathDirection] = field(init=False, default_factory=dict)

    def __post_init__(self) -> None:
        """Post init."""
        self.route_id = self.data["RouteID"]
        self.route = self.bus.routes[self.route_id]
        self.name = self.data["Name"]
        if (direction_data := self.data["Direction0"]) is not None:
            self.directions[0] = PathDirection(self.bus, direction_data, 0)
        if (direction_data := self.data["Direction1"]) is not None:
            self.directions[1] = PathDirection(self.bus, direction_data, 1)



"""Test pywmataio client for buses."""
from datetime import date, datetime

from wmataio.client import Client
from wmataio.const import TZ
from wmataio.models.area import Area


async def test_bus_apis(aioresponses):
    """Callback for aioresponse."""
    client = Client("", test_mode=True)
    await client.bus.load_data()
    assert client.bus.routes
    assert client.bus.stops

    assert len(client.bus.routes) == 390

    route = client.bus.routes["10A"]
    assert route.route_id == "10A"
    assert route.name == "10A - HUNTINGTON STA - PENTAGON"
    assert route.line_description == "Alexandria-Pentagon Line"

    assert len(client.bus.stops) == 9360

    stop = client.bus.stops["3000454"]
    assert stop.coordinates.latitude == 38.820495
    assert stop.coordinates.longitude == -76.957391
    assert stop.name == "ST BARNABAS RD + LIME ST"
    assert stop.route_ids == ["D12", "D12*5"]
    assert stop.routes == [client.bus.routes["D12"], client.bus.routes["D12*5"]]

    positions = await client.bus.get_live_positions(route=client.bus.routes["10A"])
    assert positions
    assert len(positions) == 4

    positions = await client.bus.get_live_positions(
        area=Area(2000, 38.9031442, -77.0785817)
    )
    assert positions
    assert len(positions) == 3

    positions = await client.bus.get_live_positions()
    assert positions
    assert len(positions) == 301
    position = positions[0]
    assert position.vehicle_id == "2971"
    assert position.coordinates.latitude == 38.91981887817383
    assert position.coordinates.longitude == -77.22313690185547
    assert position.deviation == -1.0
    assert position.last_update == datetime(2023, 3, 31, 22, 17, 6, tzinfo=TZ)
    assert position.trip_id == "20801060"
    assert position.route_id == "28A"
    assert position.route == client.bus.routes["28A"]
    assert position.direction_text == "WEST"
    assert position.trip_headsign == "TYSONS"
    assert position.trip_start_time == datetime(2023, 3, 31, 20, 48, 0, tzinfo=TZ)
    assert position.trip_end_time == datetime(2023, 3, 31, 22, 1, 0, tzinfo=TZ)

    incidents = await client.bus.get_bus_incidents(route=client.bus.routes["10A"])
    assert not incidents

    incidents = await client.bus.get_bus_incidents()
    assert incidents
    assert len(incidents) == 6
    incident = incidents[0]
    assert incident.incident_id == "B1D890F9-0A65-4FA4-A9AA-565B61CEC94A"
    assert incident.incident_type == "Alert"
    assert incident.route_ids_affected == ["96"]
    assert incident.routes_affected == [client.bus.routes["96"]]
    assert incident.description == (
        "Due to an incident at Tenleytown Station on the 96 route, "
        "buses are experiencing delays."
    )
    assert incident.date_updated == datetime(2023, 3, 31, 22, 13, 54, tzinfo=TZ)

    route_path = await client.bus.get_route_path(
        route=client.bus.routes["10A"], date_=date(2023, 3, 31)
    )
    assert route_path

    route_path = await client.bus.get_route_path(route=client.bus.routes["10A"])
    assert route_path
    assert route_path.route_id == "10A"
    assert route_path.route == client.bus.routes["10A"]
    assert route_path.name == "10A - PENTAGON - HUNTINGTON STA"
    assert len(route_path.directions) == 2
    assert route_path.directions[1]
    direction = route_path.directions[0]
    assert direction.trip_headsign == "PENTAGON"
    assert direction.direction == "NORTH"
    assert direction.direction_num == 0
    assert len(direction.shapes) == 184
    shape = direction.shapes[0]
    assert shape.coordinates.latitude == 38.79524
    assert shape.coordinates.longitude == -77.075059
    assert shape.sequence_number == 1
    assert len(direction.stops) == 57
    stop = direction.stops[0]
    assert stop.stop_id == "5002201"
    assert stop.name == "HUNTINGTON STATION (N) + BUS BAY B"
    assert stop.coordinates.latitude == 38.79524
    assert stop.coordinates.longitude == -77.075059
    assert stop.route_ids == ["10A"]
    assert stop.routes == [client.bus.routes["10A"]]

    route_schedule = await client.bus.get_route_schedule(
        route=client.bus.routes["10A"], date_=date(2023, 3, 31)
    )
    assert route_schedule

    route_schedule = await client.bus.get_route_schedule(route=client.bus.routes["10A"])
    assert route_schedule
    assert route_schedule.route == client.bus.routes["10A"]
    assert len(route_schedule.directions_schedules) == 2
    assert route_schedule.directions_schedules[1]
    direction_schedules = route_schedule.directions_schedules[0]
    assert len(direction_schedules) == 41
    direction_schedule = direction_schedules[0]
    assert direction_schedule.route_id == "10A"
    assert direction_schedule.route == client.bus.routes["10A"]
    assert direction_schedule.trip_direction == "NORTH"
    assert direction_schedule.trip_headsign == "PENTAGON"
    assert direction_schedule.start_time == datetime(2023, 3, 31, 4, 25, tzinfo=TZ)
    assert direction_schedule.end_time == datetime(2023, 3, 31, 5, 7, tzinfo=TZ)
    assert direction_schedule.trip_id == "30399060"
    assert len(direction_schedule.stop_times) == 57
    stop_time = direction_schedule.stop_times[0]
    assert stop_time.stop_id == "5002201"
    assert stop_time.stop == client.bus.stops["5002201"]
    assert stop_time.stop_name == "HUNTINGTON STATION (N) + BUS BAY B"
    assert stop_time.stop_sequence == 1
    assert stop_time.time == datetime(2023, 3, 31, 4, 25, tzinfo=TZ)

    next_buses = await client.bus.get_next_buses_at_stop(client.bus.stops["1002916"])
    assert len(next_buses) == 3
    next_bus = next_buses[0]
    assert next_bus.route_id == "S2"
    assert next_bus.route == client.bus.routes["S2"]
    assert next_bus.direction_text == "North to Silver Spring Station"
    assert next_bus.direction_number == 0
    assert next_bus.minutes == 0
    assert next_bus.vehicle_id == "5528"
    assert next_bus.trip_id == "33841060"

    assert (
        len(
            await client.bus.get_stop_schedule(
                client.bus.stops["1000533"], date_=date(2023, 3, 31)
            )
        )
        == 50
    )
    stop_schedule = await client.bus.get_stop_schedule(client.bus.stops["1000533"])
    assert len(stop_schedule) == 50
    stop_arrival = stop_schedule[0]
    assert stop_arrival.route_id == "V4"
    assert stop_arrival.route == client.bus.routes["V4"]
    assert stop_arrival.trip_direction_text == "EAST"
    assert stop_arrival.trip_headsign == "CAPITOL HEIGHTS STATION"
    assert stop_arrival.trip_id == "3944060"
    assert stop_arrival.direction_number == 0
    assert stop_arrival.schedule_time == datetime(2023, 3, 31, 4, 53, 28, tzinfo=TZ)
    assert stop_arrival.start_time == datetime(2023, 3, 31, 4, 47, tzinfo=TZ)
    assert stop_arrival.end_time == datetime(2023, 3, 31, 5, 24, tzinfo=TZ)



"""Exceptions for pywmataio."""


class WMATAError(BaseException):
    """Represent an error from the WMATA API."""

    message: str

    def __init__(self, message: str):
        """Initialize."""
        self.message = message



"""Standard route models for MetroRail WMATA API."""
from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Literal, TypedDict

if TYPE_CHECKING:
    from .. import MetroRail
    from .line import Line
    from .station import Station


class TrackCircuitData(TypedDict):
    """Track circuit data for MetroRail WMATA API."""

    CircuitId: int
    SeqNum: int
    StationCode: str | None


@dataclass
class TrackCircuit:
    """Track circuit."""

    bus: "MetroRail"
    data: TrackCircuitData
    circuit_id: int = field(init=False)
    sequence_number: int = field(init=False)
    station_code: str | None = field(init=False)

    def __post_init__(self) -> None:
        """Post init."""
        self.circuit_id = self.data["CircuitId"]
        self.sequence_number = self.data["SeqNum"]
        self.station_code = self.data["StationCode"]

    @property
    def station(self) -> "Station" | None:
        """Station."""
        if self.station_code is None:
            return None
        return self.bus.stations[self.station_code]


class StandardRouteData(TypedDict):
    """Standard route data for MetroRail WMATA API."""

    LineCode: str
    TrackCircuits: list[TrackCircuitData]
    TrackNum: Literal[1, 2]


@dataclass
class StandardRoute:
    """Standard route."""

    bus: "MetroRail"
    data: StandardRouteData
    line_code: str = field(init=False)
    track_circuits: list[TrackCircuit] = field(init=False)
    track_number: Literal[1, 2] = field(init=False)

    def __post_init__(self) -> None:
        """Post init."""
        self.line_code = self.data["LineCode"]
        self.track_circuits = sorted(
            [
                TrackCircuit(self.bus, track_circuit_data)
                for track_circuit_data in self.data["TrackCircuits"]
            ],
            key=lambda track_circuit: track_circuit.sequence_number,
        )
        self.track_number = self.data["TrackNum"]

    @property
    def line(self) -> "Line":
        """Return line for route."""
        return self.bus.lines[self.line_code]



"""Elevator/escalator incident models for MetroRail WMATA API."""
from __future__ import annotations

from dataclasses import dataclass, field
from datetime import datetime
from typing import TYPE_CHECKING, Literal, TypedDict

from ...const import TZ

if TYPE_CHECKING:
    from .. import MetroRail
    from .station import Station


class ElevatorAndEscalatorIncidentData(TypedDict):
    """Elevator/escalator incident data for MetroRail WMATA API."""

    DateOutOfServ: str
    DateUpdated: str
    DisplayOrder: int | None  # deprecated
    EstimatedReturnToService: str
    LocationDescription: str
    StationCode: str
    StationName: str
    SymptomCode: str | None  # deprecated
    SymptomDescription: str
    TimeOutOfService: str | None  # deprecated
    UnitName: str
    UnitStatus: str | None  # deprecated
    UnitType: Literal["ELEVATOR", "ESCALATOR"]


@dataclass
class ElevatorAndEscalatorIncident:
    """MetroRail Elevator/Escalator Incident."""

    bus: "MetroRail"
    data: ElevatorAndEscalatorIncidentData
    date_out_of_service: datetime = field(init=False)
    date_updated: datetime = field(init=False)
    estimated_return_to_service: datetime | None = field(init=False, default=None)
    location_description: str = field(init=False)
    station_code: str = field(init=False)
    station_name: str = field(init=False)
    symptom_description: str = field(init=False)
    unit_name: str = field(init=False)
    unit_type: Literal["ELEVATOR", "ESCALATOR"] = field(init=False)

    def __post_init__(self) -> None:
        """Post init."""
        self.date_out_of_service = datetime.fromisoformat(
            self.data["DateOutOfServ"]
        ).replace(tzinfo=TZ)
        self.date_updated = datetime.fromisoformat(self.data["DateUpdated"]).replace(
            tzinfo=TZ
        )
        if estimated_return_to_service := self.data["EstimatedReturnToService"]:
            self.estimated_return_to_service = datetime.fromisoformat(
                estimated_return_to_service
            ).replace(tzinfo=TZ)
        self.location_description = self.data["LocationDescription"]
        self.station_code = self.data["StationCode"]
        self.station_name = self.data["StationName"]
        self.symptom_description = self.data["SymptomDescription"]
        self.unit_name = self.data["UnitName"]
        self.unit_type = self.data["UnitType"]

    @property
    def station(self) -> "Station":
        """Return the station."""
        return self.bus.stations[self.station_code]
